# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'OperationMode',
    'Optimization',
    'DeviceType',
    'NeuralModuleFactory',
    'DeploymentFormat',
]

import random
from enum import Enum
from typing import List, Optional, Union

import numpy as np

import nemo
from nemo.core.callbacks import ActionCallback, EvaluatorCallback, NeMoCallback
from nemo.core.neural_types import NmTensor
from nemo.utils import ExpManager, logging


class DeploymentFormat(Enum):
    """Which format to use when exporting a Neural Module for deployment"""

    AUTO = 0
    PYTORCH = 1
    TORCHSCRIPT = 2
    ONNX = 3
    TRTONNX = 4
    JARVIS = 5


class OperationMode(Enum):
    """Training or Inference (Evaluation) mode"""

    training = 0
    evaluation = 1
    both = 2


class Optimization(Enum):
    """Various levels of Apex/amp Optimization.
    WARNING: This might have effect on model accuracy."""

    mxprO0 = 0
    mxprO1 = 1
    mxprO2 = 2
    mxprO3 = 3


class DeviceType(Enum):
    """Device types where Neural Modules can be placed."""

    GPU = 1
    CPU = 2
    AllGpu = 3


def _str_to_opt_level(opt_str: str) -> Optimization:
    number = int(opt_str[1:])
    if number not in Optimization._value2member_map_:
        raise ValueError(f"Unknown optimization value {opt_str}")
    return Optimization(number)


class NeuralModuleFactory(object):
    _DEFAULT = None

    """
    Neural Module Factory instance is used to create neural modules and
    trainers

    Args:
        local_rank (int): Process rank. Should be set by distributed runner
        optimization_level (Optimization): Level of optimization to use. Will
            be passed to neural modules and actions created by this factory.
        placement (DeviceType: where to place NeuralModule instances by default
        cudnn_benchmark (bool): (default False) If set to True it will use
            cudnnFind method to find the best kernels instead of using
            heuristics. If the shapes of your inputs are constant this
            should help, for various shapes it can slow things down. Give it
            few iterations to warmup if set to True. Currently only supported
            by PyTorch backend.
        random_seed (int): (default None) Sets random seed to control for
            randomness. This should be used for debugging purposes as it might
            have negative impact on performance. Can't be used when
            `cudnn_benchmark=True`.
        master_process (bool): (default True) Flag for master process
            indication
        set_default (bool): (default True) True if should set this instance as
            default factory for modules instantiating.
    """

    def __init__(
        self,
        local_rank=None,
        optimization_level=Optimization.mxprO0,
        placement=None,
        cudnn_benchmark=False,
        random_seed=None,
        set_default=True,
        log_dir=None,
        checkpoint_dir=None,
        tensorboard_dir=None,
        create_tb_writer=False,
        files_to_copy=None,
        add_time_to_log_dir=False,
    ):
        self._local_rank = local_rank
        self._global_rank = None

        if isinstance(optimization_level, str):
            optimization_level = _str_to_opt_level(optimization_level)
        self._optim_level = optimization_level

        if placement is None:
            if local_rank is not None:
                device = DeviceType.AllGpu
            else:
                device = DeviceType.GPU

            self._placement = device
        else:
            self._placement = placement

        self._world_size = 1
        broadcast_func = None
        if True:  # backend == Backend.PyTorch:
            # TODO: Move all framework specific code from this file
            import torch

            if self._placement != DeviceType.CPU:
                if not torch.cuda.is_available():
                    raise ValueError(
                        "You requested to use GPUs but CUDA is "
                        "not installed. You can try running using"
                        " CPU-only. To do this, instantiate your"
                        " factory with placement=DeviceType.CPU"
                        "\n"
                        "Note that this is slow and is not "
                        "well supported."
                    )

            torch.backends.cudnn.benchmark = cudnn_benchmark
            if random_seed is not None and cudnn_benchmark:
                raise ValueError("cudnn_benchmark can not be set to True when random_seed is not None.")
            if random_seed is not None:
                torch.backends.cudnn.deterministic = True
                torch.backends.cudnn.benchmark = False
                torch.manual_seed(random_seed)
                np.random.seed(random_seed)
                random.seed(random_seed)

            # logging.info("Random seeds")
            # logging.info("torch: %d", torch.initial_seed())
            # logging.info("numpy: %d", )
            # logging.info("random: %d", )

            if self._local_rank is not None:
                torch.distributed.init_process_group(backend="nccl", init_method="env://")

                cuda_set = True
                # Try to set cuda device. This should fail if self._local_rank
                # is greater than the number of available GPUs
                try:
                    torch.cuda.set_device(self._local_rank)
                except RuntimeError:
                    # Note in this case, all tensors are now sent to GPU 0
                    # who could crash because of OOM. Thus init_process_group()
                    # must be done before any cuda tensors are allocated
                    cuda_set = False
                cuda_set_t = torch.cuda.IntTensor([cuda_set])

                # Do an all_reduce to ensure all workers obtained a GPU
                # For the strangest reason, BAND doesn't work so I am resorting
                # to MIN.
                torch.distributed.all_reduce(cuda_set_t, op=torch.distributed.ReduceOp.MIN)
                if cuda_set_t.item() == 0:
                    raise RuntimeError(
                        "There was an error initializing distributed training."
                        " Perhaps you specified more gpus than you have "
                        "available"
                    )

                del cuda_set_t
                torch.cuda.empty_cache()
                # Remove test tensor from memory

                self._world_size = torch.distributed.get_world_size()
                self._global_rank = torch.distributed.get_rank()

                def torch_broadcast_wrapper(str_len=None, string=None, src=0):
                    """Wrapper function to broadcast string values across all
                    workers
                    """
                    # Create byte cuda torch tensor
                    if string is not None:
                        string_tensor = torch.tensor(list(string.encode()), dtype=torch.uint8).cuda()
                    else:
                        string_tensor = torch.tensor([0] * str_len, dtype=torch.uint8).cuda()
                    # Run broadcast
                    torch.distributed.broadcast(string_tensor, src)
                    # turn byte tensor back to string
                    return_string = string_tensor.cpu().numpy()
                    return_string = b''.join(return_string).decode()
                    return return_string

                broadcast_func = torch_broadcast_wrapper
        else:
            raise NotImplementedError("Only Pytorch backend is currently supported.")

        # Create ExpManager
        # if log_dir is None, only create logger
        self._exp_manager = ExpManager(
            work_dir=log_dir,
            ckpt_dir=checkpoint_dir,
            use_tb=create_tb_writer,
            tb_dir=tensorboard_dir,
            local_rank=local_rank,
            global_rank=self._global_rank,
            files_to_copy=files_to_copy,
            add_time=add_time_to_log_dir,
            exist_ok=True,
            broadcast_func=broadcast_func,
        )
        self._tb_writer = self._exp_manager.tb_writer

        # Create trainer
        self._trainer = self._get_trainer(tb_writer=self._tb_writer)

        if set_default:
            NeuralModuleFactory.set_default_factory(self)

    @classmethod
    def get_default_factory(cls):
        return cls._DEFAULT

    @classmethod
    def set_default_factory(cls, factory):
        cls._DEFAULT = factory

    @classmethod
    def reset_default_factory(cls):
        cls._DEFAULT = None

    @staticmethod
    def __name_import(name):
        components = name.split(".")
        mod = __import__(components[0])
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod

    def create_optimizer(self, optimizer, things_to_optimize, optimizer_params):
        return self._trainer.create_optimizer(
            optimizer=optimizer, things_to_optimize=things_to_optimize, optimizer_params=optimizer_params,
        )

    def train(
        self,
        tensors_to_optimize=None,
        training_graph=None,
        optimizer=None,
        optimization_params=None,
        callbacks: Optional[List[Union[ActionCallback, NeMoCallback]]] = None,
        lr_policy=None,
        batches_per_step=None,
        stop_on_nan_loss=False,
        synced_batchnorm=False,
        synced_batchnorm_groupsize=0,
        gradient_predivide=False,
        amp_max_loss_scale=2.0 ** 24,
        reset=False,
    ):
        if reset:
            self.reset_trainer()
        return self._trainer.train(
            tensors_to_optimize=tensors_to_optimize,
            training_graph=training_graph,
            optimizer=optimizer,
            optimization_params=optimization_params,
            callbacks=callbacks,
            lr_policy=lr_policy,
            batches_per_step=batches_per_step,
            stop_on_nan_loss=stop_on_nan_loss,
            synced_batchnorm=synced_batchnorm,
            synced_batchnorm_groupsize=synced_batchnorm_groupsize,
            gradient_predivide=gradient_predivide,
            amp_max_loss_scale=amp_max_loss_scale,
        )

    def eval(self, callbacks: List[EvaluatorCallback]):
        if callbacks is None or len(callbacks) == 0:
            raise ValueError(f"You need to provide at lease one evaluation" f"callback to eval")
        for callback in callbacks:
            if not isinstance(callback, EvaluatorCallback):
                raise TypeError(f"All callbacks passed to the eval action must" f"be inherited from EvaluatorCallback")
        self.train(
            tensors_to_optimize=None, optimizer='sgd', callbacks=callbacks, optimization_params={'num_epochs': 1},
        )

    def deployment_export(
        self, module, output: str, d_format: DeploymentFormat, input_example=None, output_example=None
    ):
        """Exports Neural Module instance for deployment.

        Args:
            module: neural module to export
            output (str): where export results should be saved
            d_format (DeploymentFormat): which deployment format to use
            input_example: sometimes tracing will require input examples
            output_example: Should match inference on input_example
        """
        if d_format == DeploymentFormat.JARVIS:
            logging.info("Exporting model to Jarvis.")
            module.deploy_to_jarvis(output=output)
            logging.info(f"Exported to {output}")
            return

        _inexample, _out_example = module._prepare_for_deployment()

        if input_example is not None:
            _inexample = input_example
        if output_example is not None:
            _out_example = output_example

        return self._trainer.deployment_export(
            module=module, output=output, d_format=d_format, input_example=_inexample, output_example=_out_example,
        )

    def infer(
        self,
        tensors: List[NmTensor],
        checkpoint_dir=None,
        ckpt_pattern='',
        verbose=True,
        cache=False,
        use_cache=False,
        offload_to_cpu=True,
        modules_to_restore=None,
    ):
        """Runs inference to obtain values for tensors

        Args:
            tensors (list[NmTensor]): List of NeMo tensors that we want to get
                values of.
            checkpoint_dir (str): Path to checkpoint directory. Default is None
                which does not load checkpoints.
            ckpt_pattern (str): Pattern used to check for checkpoints inside
                checkpoint_dir. Default is '' which matches any checkpoints
                inside checkpoint_dir.
            verbose (bool): Controls printing. Defaults to True.
            cache (bool): If True, cache all `tensors` and intermediate tensors
                so that future calls that have use_cache set will avoid
                computation. Defaults to False.
            use_cache (bool): Values from `tensors` will be always re-computed.
                It will re-use intermediate tensors from the DAG leading to
                `tensors`. If you want something to be re-computed, put it into
                `tensors` list. Defaults to False.
            offload_to_cpu (bool): If True, all evaluated tensors are moved to
                cpu memory after each inference batch. Defaults to True.
            modules_to_restore (list): Defaults to None, in which case all
                NMs inside callchain with weights will be restored. If
                specified only the modules inside this list will be restored.

        Returns:
            List of evaluated tensors. Each element in the list is also a list
            where each element is now a batch of tensor values.
        """
        return self._trainer.infer(
            tensors=tensors,
            checkpoint_dir=checkpoint_dir,
            ckpt_pattern=ckpt_pattern,
            verbose=verbose,
            cache=cache,
            use_cache=use_cache,
            offload_to_cpu=offload_to_cpu,
            modules_to_restore=modules_to_restore,
        )

    def clear_cache(self):
        """Helper function to clean inference cache."""
        self._trainer.clear_cache()

    def _get_trainer(self, tb_writer=None):
        if True:  # self._backend == Backend.PyTorch:
            constructor = NeuralModuleFactory.__name_import("nemo.backends.pytorch.PtActions")
            instance = constructor(
                local_rank=self._local_rank,
                global_rank=self._global_rank,
                tb_writer=tb_writer,
                optimization_level=self._optim_level,
            )
            return instance
        else:
            raise ValueError("Only PyTorch backend is currently supported.")

    def reset_trainer(self):
        del self._trainer
        self._trainer = self._get_trainer(tb_writer=self._tb_writer)

    def sync_all_processes(self, status=True):
        """ Helper function for testing that allows proccess 0 to inform all
        other processes of failures. Does nothing if not using distributed
        training. Usage example can be seen in examples/asr/jasper_an4.py

        Args:
            status (bool): Defaults to True. If any proccess passes False, it
                will trigger a graceful exit on all other processes. It is
                assumed that the process that passed False will print an error
                message on its own and exit
        """
        if self._world_size == 1:
            logging.info("sync_all_processes does nothing if there is one process")
            return
        if True:  # self._backend == Backend.PyTorch:
            import torch

            status_tensor = torch.cuda.IntTensor([status])
            torch.distributed.all_reduce(status_tensor, op=torch.distributed.ReduceOp.MIN)
            if status_tensor.item() == 0:
                logging.error("At least one process had a failure")
                if status:
                    raise ValueError(
                        f"Process with global rank {self._global_rank} entered"
                        " sync_all_processes with a passing status, but "
                        "another process indicated a failure"
                    )

    @property
    def world_size(self):
        return self._world_size

    @property
    def tb_writer(self):
        return self._tb_writer

    @property
    def placement(self):
        return self._placement

    @property
    def optim_level(self):
        return self._optim_level

    @property
    def checkpoint_dir(self):
        return self._exp_manager.ckpt_dir

    @property
    def work_dir(self):
        return self._exp_manager.work_dir

    @property
    def global_rank(self):
        return self._global_rank



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst
"""
import numpy as np
from torch.utils.data import Dataset

__all__ = ['SGDDataset']


class SGDDataset(Dataset):
    """ 
    Processes SGD dataset
    Args:
        dataset_split (str): train/dev/test
        dialogues_processor (obj): Data generator for SGD dialogues
    """

    def __init__(self, dataset_split, dialogues_processor):
        self.features = dialogues_processor.get_dialog_examples(dataset_split)

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx):
        ex = self.features[idx]
        service_id = ex.service_schema.service_id

        return (
            np.array(ex.example_id_num),
            np.array(service_id),
            np.array(ex.is_real_example, dtype=int),
            np.array(ex.utterance_ids),
            np.array(ex.utterance_segment),
            np.array(ex.utterance_mask, dtype=np.long),
            np.array(ex.categorical_slot_status),
            np.array(ex.cat_slot_status_mask),
            np.array(ex.categorical_slot_values),
            np.array(ex.cat_slot_values_mask),
            np.array(ex.noncategorical_slot_status),
            np.array(ex.noncat_slot_status_mask),
            np.array(ex.noncategorical_slot_value_start),
            np.array(ex.noncategorical_slot_value_end),
            np.array(ex.start_char_idx),  # noncat_alignment_start
            np.array(ex.end_char_idx),  # noncat_alignment_end
            np.array(ex.num_slots),  # num_requested_slots
            np.array(ex.requested_slot_status, dtype=np.float32),
            np.array(ex.requested_slot_mask),
            np.array(ex.intent_status_mask),
            np.array(ex.intent_status_labels),
        )



import math

import torch
from torch.optim import Optimizer


def _check_valid_opt_params(lr, eps, betas):
    if lr < 0:
        raise ValueError(f"Invalid learning rate: {lr}")
    if eps < 0:
        raise ValueError(f"Invalid epsilon value: {eps}")
    if not (0.0 <= betas[0] < 1.0 and 0.0 <= betas[1] < 1.0):
        raise ValueError(f"Betas have to be between 0 and 1: {betas}")


def master_params(optimizer):
    """
    Generator expression that iterates over the params owned by ``optimizer``.
    Args:
        optimizer: An optimizer previously returned from ``amp.initialize``.
    """
    for group in optimizer.param_groups:
        for p in group['params']:
            yield p


class AdamW(Optimizer):
    """Implements AdamW algorithm.
    It has been proposed in "Decoupled Weight Decay Regularization"
    (https://arxiv.org/abs/1711.05101)

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper "On the Convergence of Adam and Beyond"
    """

    def __init__(
        self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, amsgrad=False,
    ):
        _check_valid_opt_params(lr, eps, betas)
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad,)
        super(AdamW, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(AdamW, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsgrad", False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Adam does not support sparse gradients, please consider SparseAdam instead")
                amsgrad = group["amsgrad"]
                state = self.state[p]

                # State initialization
                if not state:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)
                    if amsgrad:
                        # Maintains max of all exp moving avg of squared grad
                        state["max_exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                if amsgrad:
                    max_exp_avg_sq = state["max_exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # if group['weight_decay'] != 0:
                #     grad = grad.add(group['weight_decay'], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                if amsgrad:
                    # Maintains max of all 2nd moment running avg till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group["eps"])
                else:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = group["lr"] * math.sqrt(bias_correction2) / bias_correction1

                # p.data.addcdiv_(-step_size, exp_avg, denom)
                p.data.add_(
                    -step_size, torch.mul(p.data, group["weight_decay"]).addcdiv_(1, exp_avg, denom),
                )

        return loss


class Novograd(Optimizer):
    """Implements Novograd algorithm.
    It has been proposed  in "Stochastic Gradient Methods with Layer-wise
    Adaptive Moments for Training of Deep Networks"
    (https://arxiv.org/abs/1905.11286)

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper "On the Convergence of Adam and Beyond"
    """

    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.95, 0.98),
        eps=1e-8,
        weight_decay=0,
        grad_averaging=False,
        amsgrad=False,
        luc=False,
        luc_trust=1e-3,
        luc_eps=1e-8,
    ):
        _check_valid_opt_params(lr, eps, betas)
        defaults = dict(
            lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, grad_averaging=grad_averaging, amsgrad=amsgrad,
        )
        self.luc = luc
        self.luc_trust = luc_trust
        self.luc_eps = luc_eps
        super(Novograd, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(Novograd, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsgrad", False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Sparse gradients are not supported.")
                amsgrad = group["amsgrad"]
                state = self.state[p]

                # State initialization
                if not state:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros([]).to(state["exp_avg"].device)
                    if amsgrad:
                        # Maintains max of all exp moving avg of squared grad
                        state["max_exp_avg_sq"] = torch.zeros([]).to(state["exp_avg"].device)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                if amsgrad:
                    max_exp_avg_sq = state["max_exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                norm = grad.norm().pow(2)

                if exp_avg_sq == 0:
                    exp_avg_sq.copy_(norm)
                else:
                    exp_avg_sq.mul_(beta2).add_(1 - beta2, norm)

                if amsgrad:
                    # Maintains max of all 2nd moment running avg till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group["eps"])
                else:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])

                grad.div_(denom)
                if group["weight_decay"] != 0:
                    grad.add_(group["weight_decay"], p.data)
                if group["grad_averaging"]:
                    grad.mul_(1 - beta1)
                exp_avg.mul_(beta1).add_(grad)

                if self.luc:
                    # Clip update so that updates are less than eta*weights
                    data_norm = torch.norm(p.data)
                    grad_norm = torch.norm(exp_avg.data)
                    luc_factor = self.luc_trust * data_norm / (grad_norm + self.luc_eps)
                    luc_factor = min(luc_factor, group["lr"])
                    p.data.add_(-luc_factor, exp_avg)
                else:
                    p.data.add_(-group["lr"], exp_avg)

        return loss



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from unittest import TestCase

import pytest

from nemo.backends.pytorch.tutorials.chatbot.data import loadPrepareData


class TestTutorialCornellData(TestCase):
    @pytest.mark.unit
    def test_data_preparation(self):
        datafile = "tests/data/dialog_sample.txt"
        voc, _ = loadPrepareData("cornell", datafile=datafile)
        self.assertEqual(voc.name, 'cornell')
        self.assertEqual(voc.num_words, 675)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import random

import numpy as np
from sklearn.metrics import classification_report

from nemo import logging
from nemo.collections.nlp.utils.callback_utils import list2str, plot_confusion_matrix, tensor2list

__all__ = ['eval_iter_callback', 'eval_epochs_done_callback']


def eval_iter_callback(tensors, global_vars, eval_data_layer):
    if "all_preds" not in global_vars.keys():
        global_vars["all_preds"] = []
    if "all_labels" not in global_vars.keys():
        global_vars["all_labels"] = []

    logits_lists = []
    labels_lists = []

    for kv, v in tensors.items():
        if 'logits' in kv:
            for v_tensor in v:
                for logit_tensor in v_tensor:
                    logits_lists.append(tensor2list(logit_tensor))

        if 'labels' in kv:
            for v_tensor in v:
                for label_tensor in v_tensor:
                    labels_lists.append(tensor2list(label_tensor))

    preds = list(np.argmax(np.asarray(logits_lists), 1))
    global_vars["all_preds"].extend(preds)
    global_vars["all_labels"].extend(labels_lists)


def eval_epochs_done_callback(global_vars, graph_fold):
    labels = np.asarray(global_vars['all_labels'])
    preds = np.asarray(global_vars['all_preds'])
    accuracy = sum(labels == preds) / labels.shape[0]
    logging.info(f'Accuracy: {accuracy}')

    # print predictions and labels for a small random subset of data
    sample_size = 20
    i = 0
    if preds.shape[0] > sample_size + 1:
        i = random.randint(0, preds.shape[0] - sample_size - 1)
    logging.info("Sampled preds: [%s]" % list2str(preds[i : i + sample_size]))
    logging.info("Sampled labels: [%s]" % list2str(labels[i : i + sample_size]))
    plot_confusion_matrix(labels, preds, graph_fold)
    logging.info(classification_report(labels, preds, digits=4))
    return dict({"accuracy": accuracy})



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Pytorch Dataset for training LaserTagger."""

import torch
from torch.utils.data import Dataset

__all__ = ['LaserTaggerDataset']


class LaserTaggerDataset(Dataset):
    """
    Dataset as used by the LaserTaggerDataLayer for training, validation, and inference
    pipelines.

    Args:
        preprocessed_data (str): path to preprocessed train/validation/test data
        use_t2t_decoder (bool): whether to use Autoregressive Decoder
    """

    def __init__(self, preprocessed_data, use_t2t_decoder):

        self.examples = preprocessed_data
        self.use_t2t_decoder = use_t2t_decoder

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, idx):
        input_ids = torch.Tensor(self.examples[idx].features['input_ids']).long()
        input_mask = torch.Tensor(self.examples[idx].features['input_mask'])
        segment_ids = torch.Tensor(self.examples[idx].features['segment_ids']).long()
        labels_mask = torch.Tensor(self.examples[idx].features['labels_mask'])
        labels = torch.Tensor(self.examples[idx].features['labels']).long()
        return input_ids, input_mask, segment_ids, labels, labels_mask, labels, input_mask



# Copyright (c) 2019 NVIDIA Corporation
import argparse
import os


class NemoArgParser(argparse.ArgumentParser):
    """
    This is a wrapper about argparse.

    Usage is the same as standard argparse.ArgParser:
        parser = NemoArgParser(description='Model_name')
        parser.add_argument(...)
        args = parser.parse_args()

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # NeMo arguments
        self.add_argument(
            "--local_rank", default=os.getenv('LOCAL_RANK', None), type=int, help="node rank for distributed training",
        )
        self.add_argument(
            "--amp_opt_level",
            default="O0",
            type=str,
            choices=["O0", "O1", "O2", "O3"],
            help="apex/amp precision:"
            " O0 - float32,"
            " O1 - mixed precision opt1,"
            " O2 - mixed precision opt2,"
            " O3 - float16"
            "See: https://nvidia.github.io/apex/amp.html",
        )
        self.add_argument(
            "--cudnn_benchmark",
            action="store_true",
            help="If set to True it will use cudnnFind method to"
            " find the best kernels instead of using "
            "heuristics. If the shapes of your inputs are "
            "constant this should help, for various shapes "
            "it can slow things down.",
        )
        # self.add_argument("--random_seed", default=None, type=int,
        #                   help="random seed")
        # self.add_argument("--deterministic", action="store_true",
        #                   help="whether to enable determinism")

        # Model defintion
        self.add_argument(
            "--model_config", type=str, default=None, help="model configuration file: model.yaml",
        )
        self.add_argument(
            "--train_dataset", type=str, default=None, help="training dataset path",
        )
        self.add_argument(
            "--eval_datasets", type=str, nargs="*", help="evaludation datasets paths",
        )
        self.add_argument("--batch_size", type=int, help="train batch size per GPU")
        self.add_argument(
            "--eval_batch_size", type=int, help="evaluation  batch size per GPU",
        )
        self.add_argument(
            "--eval_freq", default=1000, type=int, help="evaluation frequency, steps",
        )

        # Optimizer Choices
        self.add_argument(
            "--optimizer",
            type=str,
            choices=["sgd", "adam", "fused_adam", "adam_w", "novograd", "lamb",],
            help="optimizer",
        )
        self.add_argument("--weight_decay", type=float, default=0.0, help="weight decay")
        # self.add_argument("--momentum", type=float,
        #                   help="SGD momentum")
        # self.add_argument("--beta1", type=float,
        #                   help="Adam/AdamW/NovoGrad beta1")
        # self.add_argument("--beta2", type=float,
        #                   help="Adam/AdamW/NovoGrad beta2")

        # Optimization Arguments
        self.add_argument(
            "--num_epochs",
            type=int,
            default=None,
            help="number of epochs to train. You should specify either num_epochs or max_steps",
        )
        self.add_argument(
            "--max_steps",
            type=int,
            default=None,
            help="max number of steps to train. You should specify either num_epochs or max_steps",
        )
        self.add_argument("--lr", type=float, default=1e-3, help="base learning rate")
        self.add_argument(
            "--lr_policy", type=str, default='WarmupAnnealing', help="learning rate decay policy",
        )
        # self.add_argument("--warmup_steps", default=0, type=int,
        #                   help="number of learning rate warmup steps")
        self.add_argument(
            "--iter_per_step",
            default=1,
            type=int,
            help="number of gradients accumulation iterations per weights update step",
        )

        # Logging arguments
        self.add_argument(
            "--work_dir", default=None, type=str, help="working directory for experiment",
        )
        self.add_argument(
            "--checkpoint_dir",
            default=None,
            type=str,
            help="where to save checkpoints. If ckpt_dir is "
            "None, the default behaviour is to put it under"
            "{work_dir}/checkpoints",
        )
        self.add_argument(
            "--create_tb_writer", action="store_true", help="whether to log into Tensorboard",
        )
        self.add_argument(
            "--tensorboard_dir",
            default=None,
            type=str,
            help="If --create_tb_writer is enabled, specifies "
            "the tensorboard directory. Defaults to "
            "{work_dir}/checkpoints",
        )
        self.add_argument(
            "--checkpoint_save_freq", default=1000, type=int, help="checkpoint frequency, steps",
        )



# Copyright (c) 2019 NVIDIA Corporation
import functools
import glob
import os
from collections.abc import Iterable
from pathlib import Path
from typing import Optional

import torch
import wget

import nemo
from nemo.utils import logging

# from nemo.utils import logging


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def get_checkpoint_from_dir(module_names, cpkt_dir, ckpt_pattern='', return_steps=False):
    """ Grab all the modules with match a certain pattern in cpkt_dir
    If multiple checkpoints found, by default, use the one last created.
    """
    if not os.path.isdir(cpkt_dir):
        raise ValueError(f"{cpkt_dir} isn't a directory")

    if not isinstance(module_names, Iterable):
        module_names = [module_names]

    ckpts = []
    steps = []

    for module in module_names:
        if not isinstance(module, str):
            raise ValueError(f"Module {module} is not a string")
        if not isinstance(ckpt_pattern, str):
            raise ValueError(f"Pattern {ckpt_pattern} is not a string")

        module_ckpts = glob.glob(f'{cpkt_dir}/{module}*{ckpt_pattern}*')
        if not module_ckpts:
            raise ValueError(f'For module {module}, ' f'no file matches {ckpt_pattern} in {cpkt_dir}')

        # if multiple checkpoints match a pattern, take the latest one
        def step_from_checkpoint(checkpoint_name):
            # return step number given a checkpoint filename
            step_str = checkpoint_name.split('-')[-1].split('.')[0]
            return int(step_str)

        module_ckpt = module_ckpts[0]
        if len(module_ckpts) > 1:
            module_ckpt = max(module_ckpts, key=step_from_checkpoint)
        ckpts.append(module_ckpt)
        steps.append(step_from_checkpoint(module_ckpt))

    if return_steps:
        return ckpts, steps
    return ckpts


def _call_args_to_string(call_args):
    call_dict = {inport: value.name for inport, value in call_args.items()}
    result = "(force_pt=True,"
    counter = 0
    for key, value in call_dict.items():
        result += f"{key}={value}" if counter == 0 else f", {key}={value}"
        counter += 1
    result += ")"
    return result


def _get_instance_call_line(output_ports, instance_ref, call_str):
    result = ""
    counter = 0
    for out_port in output_ports:
        result += out_port if counter == 0 else (", " + out_port)
        counter += 1
    result += " = " + instance_ref + call_str
    return result


def get_device(local_rank: Optional[int]):
    if local_rank is not None:
        return nemo.core.DeviceType.AllGpu
    return nemo.core.DeviceType.GPU


def get_cuda_device(placement):
    """
    Converts NeMo nemo.core.DeviceType to torch.device
    Args:
        placement: nemo.core.DeviceType

    Returns:
        torch.device
    """
    gpu_devices = [nemo.core.DeviceType.GPU, nemo.core.DeviceType.AllGpu]
    return torch.device("cuda" if placement in gpu_devices else "cpu")


# def get_neural_factory(local_rank,
#                        precision,
#                        backend):
#     """
#     Helper function to create NeuralModuleFactory
#     Args:
#         local_rank:
#         precision: (nemo.core.Optimization) AMP mixed precision level
#         backend: (nemo.core.Backend) NeMo backend (defaults to Pytorch)

#     Returns:
#         An instance of the NeuralModuleFactory
#     """
#     device = nemo.utils.get_device(local_rank)
#     return nemo.core.NeuralModuleFactory(local_rank=local_rank,
#                                          optimization_level=precision,
#                                          placement=device)


def maybe_download_from_cloud(url, filename, subfolder=None, cache_dir=None, referesh_cache=False) -> str:
    """
    Helper function to download pre-trained weights from the cloud
    Args:
        url: (str) URL of storage
        filename: (str) what to download. The request will be issued to url/filename
        subfolder: (str) subfolder within cache_dir. The file will be stored in cache_dir/subfolder. Subfolder can
            be empty
        cache_dir: (str) a cache directory where to download. If not present, this function will attempt to create it.
            If None (default), then it will be $HOME/.cache/torch/NeMo
        referesh_cache: (bool) if True and cached file is present, it will delete it and re-fetch

    Returns:
        If successful - absolute local path to the downloaded file
        else - empty string
    """
    # try:
    if cache_dir is None:
        cache_location = Path.joinpath(Path.home(), '.cache/torch/NeMo')
    else:
        cache_location = cache_dir
    if subfolder is not None:
        destination = Path.joinpath(cache_location, subfolder)
    else:
        destination = cache_location

    if not os.path.exists(destination):
        os.makedirs(destination, exist_ok=True)

    destination_file = Path.joinpath(destination, filename)

    if os.path.exists(destination_file):
        logging.info(f"Found existing object {destination_file}.")
        if referesh_cache:
            logging.info("Asked to refresh the cache.")
            logging.info(f"Deleting file: {destination_file}")
            os.remove(destination_file)
        else:
            logging.info(f"Re-using file from: {destination_file}")
            return str(destination_file)
    # download file
    wget_uri = url + filename
    logging.info(f"Downloading from: {wget_uri} to {str(destination_file)}")
    wget.download(wget_uri, str(destination_file))
    if os.path.exists(destination_file):
        return destination_file
    else:
        return ""



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# =============================================================================
# Copyright 2019 Salesforce Research.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# =============================================================================


import random

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn as nn

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs

__all__ = ['TRADEGenerator']


class TRADEGenerator(TrainableNM):
    """
    The generator module for state tracking model TRADE
    Args:
        vocab (Vocab): an instance of Vocab containing the vocabularey
        embeddings (Tensor): word embedding matrix
        hid_size (int): hidden size of the GRU decoder
        dropout (float): dropout of the GRU
        slots (list): list of slots
        nb_gate (int): number of gates
        teacher_forcing (float): 0.5
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            'encoder_hidden': NeuralType(('B', 'T', 'D'), ChannelType()),
            'encoder_outputs': NeuralType(('B', 'T', 'D'), ChannelType()),
            'dialog_ids': NeuralType(('B', 'T'), elements_type=TokenIndex()),
            'dialog_lens': NeuralType(tuple('B'), elements_type=Length()),
            'targets': NeuralType(('B', 'D', 'T'), LabelsType(), optional=True),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        point_outputs: outputs of the generator
        gate_outputs: outputs of gating heads
        """
        return {
            'point_outputs': NeuralType(('B', 'T', 'D', 'D'), LogitsType()),
            'gate_outputs': NeuralType(('B', 'D', 'D'), LogitsType()),
        }

    def __init__(self, vocab, embeddings, hid_size, dropout, slots, nb_gate, teacher_forcing=0.5, max_res_len=10):
        super().__init__()
        self.vocab_size = len(vocab)
        self.vocab = vocab
        self.embedding = embeddings
        self.dropout = nn.Dropout(dropout)
        self.rnn = nn.GRU(hid_size, hid_size, dropout=dropout, batch_first=True)
        self.nb_gate = nb_gate
        self.hidden_size = hid_size
        self.w_ratio = nn.Linear(3 * hid_size, 1)
        self.w_gate = nn.Linear(hid_size, nb_gate)
        self.softmax = nn.Softmax(dim=1)
        self.sigmoid = nn.Sigmoid()
        self.slots = slots
        self.teacher_forcing = teacher_forcing
        # max_res_len is used in evaluation mode or when targets are not provided
        self.max_res_len = max_res_len

        self._slots_split_to_index()
        self.slot_emb = nn.Embedding(len(self.slot_w2i), hid_size)
        self.slot_emb.weight.data.normal_(0, 0.1)
        self.to(self._device)

    def _slots_split_to_index(self):
        split_slots = [slot.split('-') for slot in self.slots]
        domains = [split_slot[0] for split_slot in split_slots]
        slots = [split_slot[1] for split_slot in split_slots]
        split_slots = list({s: 0 for s in sum(split_slots, [])})
        self.slot_w2i = {split_slots[i]: i for i in range(len(split_slots))}
        self.domain_idx = torch.tensor([self.slot_w2i[domain] for domain in domains], device=self._device)
        self.subslot_idx = torch.tensor([self.slot_w2i[slot] for slot in slots], device=self._device)

    def forward(self, encoder_hidden, encoder_outputs, dialog_ids, dialog_lens, targets=None):
        if (not self.training) or (random.random() > self.teacher_forcing):
            use_teacher_forcing = False
        else:
            use_teacher_forcing = True

        batch_size = encoder_hidden.shape[0]

        if isinstance(targets, torch.Tensor):
            max_res_len = targets.shape[2]
            targets = targets.transpose(0, 1)
        else:
            max_res_len = self.max_res_len

        all_point_outputs = torch.zeros(len(self.slots), batch_size, max_res_len, self.vocab_size, device=self._device)
        all_gate_outputs = torch.zeros(len(self.slots), batch_size, self.nb_gate, device=self._device)

        domain_emb = self.slot_emb(self.domain_idx).to(self._device)
        subslot_emb = self.slot_emb(self.subslot_idx).to(self._device)
        slot_emb = domain_emb + subslot_emb
        slot_emb = slot_emb.unsqueeze(1)
        slot_emb = slot_emb.repeat(1, batch_size, 1)
        decoder_input = self.dropout(slot_emb).view(-1, self.hidden_size)
        hidden = encoder_hidden[:, 0:1, :].transpose(0, 1).repeat(len(self.slots), 1, 1)

        hidden = hidden.view(-1, self.hidden_size).unsqueeze(0)

        enc_len = dialog_lens.repeat(len(self.slots))

        maxlen = encoder_outputs.size(1)
        padding_mask_bool = ~(torch.arange(maxlen, device=self._device)[None, :] <= enc_len[:, None])
        padding_mask = torch.zeros_like(padding_mask_bool, dtype=encoder_outputs.dtype, device=self._device)
        padding_mask.masked_fill_(mask=padding_mask_bool, value=-np.inf)

        for wi in range(max_res_len):
            dec_state, hidden = self.rnn(decoder_input.unsqueeze(1), hidden)

            enc_out = encoder_outputs.repeat(len(self.slots), 1, 1)
            context_vec, logits, prob = TRADEGenerator.attend(enc_out, hidden.squeeze(0), padding_mask)

            if wi == 0:
                all_gate_outputs = torch.reshape(self.w_gate(context_vec), all_gate_outputs.size())

            p_vocab = TRADEGenerator.attend_vocab(self.embedding.weight, hidden.squeeze(0))
            p_gen_vec = torch.cat([dec_state.squeeze(1), context_vec, decoder_input], -1)
            vocab_pointer_switches = self.sigmoid(self.w_ratio(p_gen_vec))
            p_context_ptr = torch.zeros(p_vocab.size(), device=self._device)

            p_context_ptr.scatter_add_(1, dialog_ids.repeat(len(self.slots), 1), prob)

            final_p_vocab = (1 - vocab_pointer_switches).expand_as(
                p_context_ptr
            ) * p_context_ptr + vocab_pointer_switches.expand_as(p_context_ptr) * p_vocab
            pred_word = torch.argmax(final_p_vocab, dim=1)

            all_point_outputs[:, :, wi, :] = torch.reshape(
                final_p_vocab, (len(self.slots), batch_size, self.vocab_size)
            )

            if use_teacher_forcing and isinstance(targets, torch.Tensor):
                decoder_input = self.embedding(torch.flatten(targets[:, :, wi]))
            else:
                decoder_input = self.embedding(pred_word)

            decoder_input = decoder_input.to(self._device)
        all_point_outputs = all_point_outputs.transpose(0, 1).contiguous()
        all_gate_outputs = all_gate_outputs.transpose(0, 1).contiguous()
        return all_point_outputs, all_gate_outputs

    @staticmethod
    def attend(seq, cond, padding_mask):
        scores_ = cond.unsqueeze(1).expand_as(seq).mul(seq).sum(2)
        scores_ = scores_ + padding_mask
        scores = F.softmax(scores_, dim=1)
        context = scores.unsqueeze(2).expand_as(seq).mul(seq).sum(1)
        return context, scores_, scores

    @staticmethod
    def attend_vocab(seq, cond):
        scores_ = cond.matmul(seq.transpose(1, 0))
        scores = F.softmax(scores_, dim=1)
        return scores



# Copyright (c) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
This file contains neural modules responsible for preprocessing audio data.
"""
__all__ = [
    'AudioPreprocessing',
    'AudioPreprocessor',
    'AudioToMFCCPreprocessor',
    'AudioToMelSpectrogramPreprocessor',
    'AudioToSpectrogramPreprocessor',
    'CropOrPadSpectrogramAugmentation',
    'MultiplyBatch',
    'SpectrogramAugmentation',
    'TimeStretchAugmentation',
]

import math
from abc import abstractmethod

import numpy as np
import torch
from packaging import version

from .parts.features import FilterbankFeatures
from .parts.spectr_augment import SpecAugment, SpecCutout
from nemo.backends.pytorch import NonTrainableNM
from nemo.core import Optimization
from nemo.core.neural_types import *
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs

try:
    import torchaudio
    import torchaudio.transforms
    import torchaudio.functional

    TORCHAUDIO_VERSION = version.parse(torchaudio.__version__)
    TORCHAUDIO_VERSION_MIN = version.parse('0.5')

    HAVE_TORCHAUDIO = True
except ModuleNotFoundError:
    HAVE_TORCHAUDIO = False
    logging.warning('Could not import torchaudio. Some features might not work.')

try:
    from apex import amp
except (AttributeError, ModuleNotFoundError) as e:
    logging.warning("Unable to import APEX. Mixed precision and distributed training will not work.")


class AudioPreprocessor(NonTrainableNM):
    """
    A base class for Neural Modules that performs audio preprocessing,
    transforming the wav files to features.
    """

    def __init__(self, win_length, hop_length):
        super().__init__()

        self.win_length = win_length
        self.hop_length = hop_length

        self.disable_casts = self._opt_level == Optimization.mxprO1

        self.torch_windows = {
            'hann': torch.hann_window,
            'hamming': torch.hamming_window,
            'blackman': torch.blackman_window,
            'bartlett': torch.bartlett_window,
            'ones': torch.ones,
            None: torch.ones,
        }

    @torch.no_grad()
    def forward(self, input_signal, length):
        if self.disable_casts:
            with amp.disable_casts():
                processed_signal = self.get_features(input_signal.to(torch.float), length)
        else:
            processed_signal = self.get_features(input_signal, length)

        processed_length = self.get_seq_len(length.float())

        return processed_signal, processed_length

    @abstractmethod
    def get_features(self, input_signal, length):
        # Called by forward(). Subclasses should implement this.
        pass

    def get_seq_len(self, length):
        # Called by forward()
        return torch.ceil(length / self.hop_length).to(dtype=torch.long)


class AudioToSpectrogramPreprocessor(AudioPreprocessor):
    """Preprocessor that converts wavs to spectrograms.
    Uses torchaudio's Spectrogram class as a featurizer.

    Args:
        sample_rate (int): Sample rate of the input audio data.
            Defaults to 16000
        window_size (float): Size of window for fft in seconds
            Defaults to 0.02
        window_stride (float): Stride of window for fft in seconds
            Defaults to 0.01
        n_window_size (int): Size of window for fft in samples
            Defaults to None. Use one of window_size or n_window_size.
        n_window_stride (int): Stride of window for fft in samples
            Defaults to None. Use one of window_stride or n_window_stride.
        n_fft (int): Length of FT window. If None, it uses the smallest power
            of 2 that is larger than n_window_size.
            Defaults to None
        window (str): Windowing function for fft. can be one of ['hann',
            'hamming', 'blackman', 'bartlett', 'none', 'null']
            Defaults to "hann"
        normalized (bool): Whether to normalize by magnitude after stft
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "input_signal": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "length": NeuralType({0: AxisType(BatchTag)}),
            "input_signal": NeuralType(('B', 'T'), AudioSignal(freq=self._sample_rate)),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "processed_signal": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(ProcessedTimeTag),}
            # ),
            # "processed_length": NeuralType({0: AxisType(BatchTag)}),
            "processed_signal": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "processed_length": NeuralType(tuple('B'), LengthsType()),
        }

    def __init__(
        self,
        sample_rate=16000,
        window_size=0.02,
        window_stride=0.01,
        n_window_size=None,
        n_window_stride=None,
        n_fft=None,
        window="hann",
        normalized=True,
    ):
        self._sample_rate = sample_rate
        if not HAVE_TORCHAUDIO:
            raise ModuleNotFoundError(
                "torchaudio is not installed but is necessary for "
                "AudioToSpectrogramPreprocessor. We recommend you try "
                "building it from source for the PyTorch version you have."
            )
        if window_size and n_window_size:
            raise ValueError(f"{self} received both window_size and " f"n_window_size. Only one should be specified.")
        if window_stride and n_window_stride:
            raise ValueError(
                f"{self} received both window_stride and " f"n_window_stride. Only one should be specified."
            )
        if window_size:
            n_window_size = int(window_size * self._sample_rate)
        if window_stride:
            n_window_stride = int(window_stride * self._sample_rate)

        super().__init__(n_window_size, n_window_stride)

        self.win_length = n_window_size
        self.hop_length = n_window_stride

        self.n_fft = n_fft or 2 ** math.ceil(math.log2(self.win_length))

        # Set window_fn. None defaults to torch.ones.
        window_fn = self.torch_windows.get(window, None)
        if window_fn is None:
            raise ValueError(
                f"Window argument for AudioProcessor is invalid: {window}."
                f"For no window function, use 'ones' or None."
            )

        # Create featurizer.
        # Calls torch.stft under the hood, and is hard-coded to use center=True
        self.featurizer = torchaudio.transforms.Spectrogram(
            n_fft=self.n_fft,
            win_length=self.win_length,
            hop_length=self.hop_length,
            window_fn=window_fn,
            normalized=normalized,
        )
        self.featurizer.to(self._device)

    def get_features(self, input_signal, length):
        return self.featurizer(input_signal)

    @property
    def sample_rate(self):
        return self._sample_rate


class AudioToMelSpectrogramPreprocessor(AudioPreprocessor):
    """Featurizer that converts wavs to mel spectrograms.
    We don't use torchaudio's implementation here because the original
    implementation is not the same, so for the sake of backwards-compatibility
    this will use the old FilterbankFeatures for now.

    Args:
        sample_rate (int): Sample rate of the input audio data.
            Defaults to 16000
        window_size (float): Size of window for fft in seconds
            Defaults to 0.02
        window_stride (float): Stride of window for fft in seconds
            Defaults to 0.01
        n_window_size (int): Size of window for fft in samples
            Defaults to None. Use one of window_size or n_window_size.
        n_window_stride (int): Stride of window for fft in samples
            Defaults to None. Use one of window_stride or n_window_stride.
        window (str): Windowing function for fft. can be one of ['hann',
            'hamming', 'blackman', 'bartlett']
            Defaults to "hann"
        normalize (str): Can be one of ['per_feature', 'all_features']; all
            other options disable feature normalization. 'all_features'
            normalizes the entire spectrogram to be mean 0 with std 1.
            'pre_features' normalizes per channel / freq instead.
            Defaults to "per_feature"
        n_fft (int): Length of FT window. If None, it uses the smallest power
            of 2 that is larger than n_window_size.
            Defaults to None
        preemph (float): Amount of pre emphasis to add to audio. Can be
            disabled by passing None.
            Defaults to 0.97
        features (int): Number of mel spectrogram freq bins to output.
            Defaults to 64
        lowfreq (int): Lower bound on mel basis in Hz.
            Defaults to 0
        highfreq  (int): Lower bound on mel basis in Hz.
            Defaults to None
        log (bool): Log features.
            Defaults to True
        log_zero_guard_type(str): Need to avoid taking the log of zero. There
            are two options: "add" or "clamp".
            Defaults to "add".
        log_zero_guard_value(float, or str): Add or clamp requires the number
            to add with or clamp to. log_zero_guard_value can either be a float
            or "tiny" or "eps". torch.finfo is used if "tiny" or "eps" is
            passed.
            Defaults to 2**-24.
        dither (float): Amount of white-noise dithering.
            Defaults to 1e-5
        pad_to (int): Ensures that the output size of the time dimension is
            a multiple of pad_to.
            Defaults to 16
        frame_splicing (int): Defaults to 1
        stft_conv (bool): If True, uses pytorch_stft and convolutions. If
            False, uses torch.stft.
            Defaults to False
        pad_value (float): The value that shorter mels are padded with.
            Defaults to 0
        mag_power (float): The power that the linear spectrogram is raised to
            prior to multiplication with mel basis.
            Defaults to 2 for a power spec
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "input_signal": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "length": NeuralType({0: AxisType(BatchTag)}),
            "input_signal": NeuralType(('B', 'T'), AudioSignal(freq=self._sample_rate)),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        processed_signal:

            0: AxisType(BatchTag)

            1: AxisType(MelSpectrogramSignalTag)

            2: AxisType(ProcessedTimeTag)

        processed_length:

            0: AxisType(BatchTag)

        """
        return {
            # "processed_signal": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(MelSpectrogramSignalTag), 2: AxisType(ProcessedTimeTag),}
            # ),
            # "processed_length": NeuralType({0: AxisType(BatchTag)}),
            "processed_signal": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            "processed_length": NeuralType(tuple('B'), LengthsType()),
        }

    def __init__(
        self,
        sample_rate=16000,
        window_size=0.02,
        window_stride=0.01,
        n_window_size=None,
        n_window_stride=None,
        window="hann",
        normalize="per_feature",
        n_fft=None,
        preemph=0.97,
        features=64,
        lowfreq=0,
        highfreq=None,
        log=True,
        log_zero_guard_type="add",
        log_zero_guard_value=2 ** -24,
        dither=1e-5,
        pad_to=16,
        frame_splicing=1,
        stft_conv=False,
        pad_value=0,
        mag_power=2.0,
    ):
        self._sample_rate = sample_rate
        if window_size and n_window_size:
            raise ValueError(f"{self} received both window_size and " f"n_window_size. Only one should be specified.")
        if window_stride and n_window_stride:
            raise ValueError(
                f"{self} received both window_stride and " f"n_window_stride. Only one should be specified."
            )
        if window_size:
            n_window_size = int(window_size * self._sample_rate)
        if window_stride:
            n_window_stride = int(window_stride * self._sample_rate)

        super().__init__(n_window_size, n_window_stride)

        self.featurizer = FilterbankFeatures(
            sample_rate=self._sample_rate,
            n_window_size=n_window_size,
            n_window_stride=n_window_stride,
            window=window,
            normalize=normalize,
            n_fft=n_fft,
            preemph=preemph,
            nfilt=features,
            lowfreq=lowfreq,
            highfreq=highfreq,
            log=log,
            log_zero_guard_type=log_zero_guard_type,
            log_zero_guard_value=log_zero_guard_value,
            dither=dither,
            pad_to=pad_to,
            frame_splicing=frame_splicing,
            stft_conv=stft_conv,
            pad_value=pad_value,
            mag_power=mag_power,
        )
        self.featurizer.to(self._device)

    def get_features(self, input_signal, length):
        return self.featurizer(input_signal, length)

    def get_seq_len(self, seq_len):
        return self.featurizer.get_seq_len(seq_len)

    @property
    def filter_banks(self):
        return self.featurizer.filter_banks


class AudioToMFCCPreprocessor(AudioPreprocessor):
    """Preprocessor that converts wavs to MFCCs.
    Uses torchaudio.transforms.MFCC.

    Args:
        sample_rate: The sample rate of the audio.
            Defaults to 16000.
        window_size: Size of window for fft in seconds. Used to calculate the
            win_length arg for mel spectrogram.
            Defaults to 0.02
        window_stride: Stride of window for fft in seconds. Used to caculate
            the hop_length arg for mel spect.
            Defaults to 0.01
        n_window_size: Size of window for fft in samples
            Defaults to None. Use one of window_size or n_window_size.
        n_window_stride: Stride of window for fft in samples
            Defaults to None. Use one of window_stride or n_window_stride.
        window: Windowing function for fft. can be one of ['hann',
            'hamming', 'blackman', 'bartlett', 'none', 'null'].
            Defaults to 'hann'
        n_fft: Length of FT window. If None, it uses the smallest power of 2
            that is larger than n_window_size.
            Defaults to None
        lowfreq (int): Lower bound on mel basis in Hz.
            Defaults to 0
        highfreq  (int): Lower bound on mel basis in Hz.
            Defaults to None
        n_mels: Number of mel filterbanks.
            Defaults to 64
        n_mfcc: Number of coefficients to retain
            Defaults to 64
        dct_type: Type of discrete cosine transform to use
        norm: Type of norm to use
        log: Whether to use log-mel spectrograms instead of db-scaled.
            Defaults to True.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "input_signal": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "length": NeuralType({0: AxisType(BatchTag)}),
            "input_signal": NeuralType(('B', 'T'), AudioSignal(freq=self._sample_rate)),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "processed_signal": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(MFCCSignalTag), 2: AxisType(ProcessedTimeTag),}
            # ),
            # "processed_length": NeuralType({0: AxisType(BatchTag)}),
            "processed_signal": NeuralType(('B', 'D', 'T'), MFCCSpectrogramType()),
            "processed_length": NeuralType(tuple('B'), LengthsType()),
        }

    def __init__(
        self,
        sample_rate=16000,
        window_size=0.02,
        window_stride=0.01,
        n_window_size=None,
        n_window_stride=None,
        window='hann',
        n_fft=None,
        lowfreq=0.0,
        highfreq=None,
        n_mels=64,
        n_mfcc=64,
        dct_type=2,
        norm='ortho',
        log=True,
    ):
        self._sample_rate = sample_rate
        if not HAVE_TORCHAUDIO:
            raise ModuleNotFoundError(
                "torchaudio is not installed but is necessary for "
                "AudioToMFCCPreprocessor. We recommend you try "
                "building it from source for the PyTorch version you have."
            )
        if window_size and n_window_size:
            raise ValueError(f"{self} received both window_size and " f"n_window_size. Only one should be specified.")
        if window_stride and n_window_stride:
            raise ValueError(
                f"{self} received both window_stride and " f"n_window_stride. Only one should be specified."
            )
        # Get win_length (n_window_size) and hop_length (n_window_stride)
        if window_size:
            n_window_size = int(window_size * self._sample_rate)
        if window_stride:
            n_window_stride = int(window_stride * self._sample_rate)

        super().__init__(n_window_size, n_window_stride)

        mel_kwargs = {}

        mel_kwargs['f_min'] = lowfreq
        mel_kwargs['f_max'] = highfreq
        mel_kwargs['n_mels'] = n_mels

        mel_kwargs['n_fft'] = n_fft or 2 ** math.ceil(math.log2(n_window_size))

        mel_kwargs['win_length'] = n_window_size
        mel_kwargs['hop_length'] = n_window_stride

        # Set window_fn. None defaults to torch.ones.
        window_fn = self.torch_windows.get(window, None)
        if window_fn is None:
            raise ValueError(
                f"Window argument for AudioProcessor is invalid: {window}."
                f"For no window function, use 'ones' or None."
            )
        mel_kwargs['window_fn'] = window_fn

        # Use torchaudio's implementation of MFCCs as featurizer
        self.featurizer = torchaudio.transforms.MFCC(
            sample_rate=self._sample_rate,
            n_mfcc=n_mfcc,
            dct_type=dct_type,
            norm=norm,
            log_mels=log,
            melkwargs=mel_kwargs,
        )
        self.featurizer.to(self._device)

    def get_features(self, input_signal, length):
        return self.featurizer(input_signal)


class SpectrogramAugmentation(NonTrainableNM):
    """
    Performs time and freq cuts in one of two ways.

    SpecAugment zeroes out vertical and horizontal sections as described in
    SpecAugment (https://arxiv.org/abs/1904.08779). Arguments for use with
    SpecAugment are `freq_masks`, `time_masks`, `freq_width`, and `time_width`.

    SpecCutout zeroes out rectangulars as described in Cutout
    (https://arxiv.org/abs/1708.04552). Arguments for use with Cutout are
    `rect_masks`, `rect_freq`, and `rect_time`.

    Args:
        freq_masks (int): how many frequency segments should be cut.
            Defaults to 0.
        time_masks (int): how many time segments should be cut
            Defaults to 0.
        freq_width (int): maximum number of frequencies to be cut in one
            segment.
            Defaults to 10.
        time_width (int): maximum number of time steps to be cut in one
            segment
            Defaults to 10.
        rect_masks (int): how many rectangular masks should be cut
            Defaults to 0.
        rect_freq (int): maximum size of cut rectangles along the frequency
            dimension
            Defaults to 5.
        rect_time (int): maximum size of cut rectangles along the time
            dimension
            Defaults to 25.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "input_spec": NeuralType({0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(
            # TimeTag),})
            "input_spec": NeuralType(('B', 'D', 'T'), SpectrogramType())
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "augmented_spec": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(ProcessedTimeTag),}
            # )
            "augmented_spec": NeuralType(('B', 'D', 'T'), SpectrogramType())
        }

    def __init__(
        self,
        freq_masks=0,
        time_masks=0,
        freq_width=10,
        time_width=10,
        rect_masks=0,
        rect_time=5,
        rect_freq=20,
        rng=None,
    ):
        super().__init__()

        if rect_masks > 0:
            self.spec_cutout = SpecCutout(rect_masks=rect_masks, rect_time=rect_time, rect_freq=rect_freq, rng=rng,)
            self.spec_cutout.to(self._device)
        else:
            self.spec_cutout = lambda x: x

        if freq_masks + time_masks > 0:
            self.spec_augment = SpecAugment(
                freq_masks=freq_masks, time_masks=time_masks, freq_width=freq_width, time_width=time_width, rng=rng,
            )
            self.spec_augment.to(self._device)
        else:
            self.spec_augment = lambda x: x

    def forward(self, input_spec):
        augmented_spec = self.spec_cutout(input_spec)
        augmented_spec = self.spec_augment(augmented_spec)
        return augmented_spec


class MultiplyBatch(NonTrainableNM):
    """
    Augmentation that repeats each element in a batch.
    Other augmentations can be applied afterwards.

    Args:
        mult_batch (int): number of repeats
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "in_x": NeuralType({0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(TimeTag),}),
            # "in_x_len": NeuralType({0: AxisType(BatchTag)}),
            # "in_y": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "in_y_len": NeuralType({0: AxisType(BatchTag)}),
            "in_x": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "in_x_len": NeuralType(tuple('B'), LengthsType()),
            "in_y": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "in_y_len": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "out_x": NeuralType({0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(TimeTag),}),
            # "out_x_len": NeuralType({0: AxisType(BatchTag)}),
            # "out_y": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "out_y_len": NeuralType({0: AxisType(BatchTag)}),
            "out_x": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "out_x_len": NeuralType(tuple('B'), LengthsType()),
            "out_y": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "out_y_len": NeuralType(tuple('B'), LengthsType()),
        }

    def __init__(self, mult_batch=1):
        super().__init__()
        self.mult = mult_batch

    @torch.no_grad()
    def forward(self, in_x, in_x_len, in_y, in_y_len):
        out_x = in_x.repeat(self.mult, 1, 1)
        out_y = in_y.repeat(self.mult, 1)
        out_x_len = in_x_len.repeat(self.mult)
        out_y_len = in_y_len.repeat(self.mult)

        return out_x, out_x_len, out_y, out_y_len


class CropOrPadSpectrogramAugmentation(NonTrainableNM):
    """
    Pad or Crop the incoming Spectrogram to a certain shape.

    Args:
        audio_length (int): the final number of timesteps that is required.
            The signal will be either padded or cropped temporally to this
            size.
    """

    def __init__(self, audio_length, **kwargs):
        super(CropOrPadSpectrogramAugmentation, self).__init__()
        self.audio_length = audio_length

    @torch.no_grad()
    def forward(self, input_signal, length):
        image = input_signal
        num_images = image.shape[0]

        audio_length = self.audio_length
        image_len = image.shape[-1]

        # Crop long signal
        if image_len > audio_length:  # randomly slice
            cutout_images = []
            offset = torch.randint(low=0, high=image_len - audio_length + 1, size=[num_images])

            # TODO: Look into advanced broadcasting to speed up section
            for idx, offset in enumerate(offset):
                cutout_images.append(image[idx : idx + 1, :, offset : offset + audio_length])

            image = torch.cat(cutout_images, dim=0)
            del cutout_images

        else:  # symmetrically pad short signal with zeros
            pad_left = (audio_length - image_len) // 2
            pad_right = (audio_length - image_len) // 2

            if (audio_length - image_len) % 2 == 1:
                pad_right += 1

            image = torch.nn.functional.pad(image, [pad_left, pad_right], mode="constant", value=0)

        # Replace dynamic length sequences with static number of timesteps
        length = (length * 0) + audio_length

        return image, length

    @property
    def input_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "input_signal": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(ProcessedTimeTag), }
            # ),
            # "length": NeuralType({0: AxisType(BatchTag)}),
            "input_signal": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "processed_signal": NeuralType(
            #     {0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(ProcessedTimeTag), }
            # ),
            # "processed_length": NeuralType({0: AxisType(BatchTag)}),
            "processed_signal": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "processed_length": NeuralType(tuple('B'), LengthsType()),
        }


class TimeStretchAugmentation(NonTrainableNM):
    def __init__(
        self,
        sample_rate: int,
        probability: float,
        min_speed_rate: float = 0.9,
        max_speed_rate: float = 1.1,
        num_rates: int = 5,
        n_fft: int = 512,
    ):
        """
        Time-stretch a batch of audio series by a fixed rate while preserving pitch.

        Note that while the speed rate is sampled independently for every batch,
        all samples of that batch will be augmented by the same speed rate.

        Note:
        This is a simplified implementation, intended primarily for reference and pedagogical purposes.
        It makes no attempt to handle transients, and is likely to produce audible artifacts.

        Args:
            sample_rate: Sampling rate.
            probability: Float value declaring chance of the input being augmented.
                Must be a float value in the range [0, 1].
            min_speed_rate: Minimum sampling rate modifier.
            max_speed_rate: Maximum sampling rate modifier.
            num_rates: Number of discrete rates to allow. Can be a positive or negative
                integer.
                If a positive integer greater than 0 is provided, the range of
                speed rates will be discretized into `num_rates` values.
                If a negative integer or 0 is provided, the full range of speed rates
                will be sampled uniformly.
                Note: If a positive integer is provided and the resultant discretized
                range of rates contains the value '1.0', then those samples with rate=1.0,
                will not be augmented at all and simply skipped. This is to avoid unnecessary
                augmentation and increase computation time. Effective augmentation chance
                in such a case is = `prob * (num_rates - 1 / num_rates) * 100`% chance
                where `prob` is the global probability of a sample being augmented.
            n_fft: Number of fft filters to be computed.
        """
        super(TimeStretchAugmentation, self).__init__()

        if probability > 1.0 or probability < 0.0:
            raise ValueError("`probability` must be between 0 and 1")

        if not HAVE_TORCHAUDIO:
            raise ModuleNotFoundError(
                "torchaudio is not installed but is necessary for "
                "TimeStretchAugmentation. We recommend you try "
                "installing it from conda for the PyTorch version you have."
            )

        # Check torchaudio version; inform user of potential issue
        if TORCHAUDIO_VERSION < TORCHAUDIO_VERSION_MIN:
            logging.error(
                "Current installed version of `torchaudio` %s is less than the recommended minimum "
                "version of %s. Please note that this may cause deadlocks when using distributed "
                "data parallel training. Please follow the instructions at https://github.com/pytorch/audio "
                "to update torchaudio.",
                str(TORCHAUDIO_VERSION),
                str(TORCHAUDIO_VERSION_MIN),
            )

        min_rate = min(min_speed_rate, max_speed_rate)
        if min_rate < 0.0:
            raise ValueError("Minimum sampling rate modifier must be > 0.")

        self._sample_rate = sample_rate
        self.probability = float(probability)
        self.min_rate = float(min_speed_rate)
        self.max_rate = float(max_speed_rate)
        self.num_rates = num_rates
        if num_rates > 0:
            self._rates = np.linspace(min_speed_rate, max_speed_rate, num_rates)
        self._rng = np.random.RandomState()

        self._n_fft = n_fft
        self._hop_length = n_fft // 2
        self._stft_window = torch.hann_window(self._n_fft, periodic=True, device=self._device)
        self._phi_advance = torch.linspace(0, np.pi * self._hop_length, self._hop_length + 1, device=self._device)
        self._phi_advance = self._phi_advance.view(-1, 1)

    @torch.no_grad()
    def forward(self, input_signal, length):
        proba = self._rng.uniform(0.0, 1.0)

        if proba > self.probability:
            return input_signal, length

        # Select speed rate either from choice or random sample
        if self.num_rates < 0:
            speed_rate = self._rng.uniform(self.min_rate, self.max_rate)
        else:
            speed_rate = np.random.choice(self._rates)

        # Skip perturbation in case of identity speed rate
        if speed_rate == 1.0:
            return input_signal, length

        features = self._stft(input_signal, self._n_fft, self._hop_length)
        features = self._phase_vocoder(features, speed_rate)

        # Predict the length of y_stretch
        len_stretch = int(round(input_signal.shape[1] / speed_rate))

        audio = self._istft(features, len_stretch)

        length = (length * speed_rate).type(torch.long)

        return audio, length

    def _stft(self, data: torch.Tensor, n_fft: int, hop_length: int):
        win_length = n_fft
        window = self._stft_window

        stft = torch.stft(
            data,
            n_fft=n_fft,
            hop_length=hop_length,
            win_length=win_length,
            window=window,
            center=True,
            pad_mode='reflect',
            normalized=False,
        )
        return stft

    def _phase_vocoder(self, data: torch.Tensor, rate: float):
        data_stretch = torchaudio.functional.phase_vocoder(data, rate, self._phi_advance)
        return data_stretch

    def _istft(self, data: torch.Tensor, len_stretch: int):
        n_fft = 2 * (data.shape[1] - 1)
        hop_length = self._hop_length
        win_length = n_fft
        window = self._stft_window

        audio = torchaudio.functional.istft(
            data,
            n_fft,
            hop_length,
            win_length,
            window=window,
            center=True,
            pad_mode='reflect',
            normalized=False,
            length=len_stretch,
        )

        return audio

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "input_signal": NeuralType(('B', 'T'), AudioSignal(freq=self._sample_rate)),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            "processed_signal": NeuralType(('B', 'T'), AudioSignal(freq=self._sample_rate)),
            "processed_length": NeuralType(tuple('B'), LengthsType()),
        }


def AudioPreprocessing(*args, **kwargs):
    raise NotImplementedError(
        "AudioPreprocessing has been deprecated and replaced by: "
        "AudioToMFCCPreprocessor, AudioToMelSpectrogramPreprocessor, and "
        "AudioToSpectrogramPreprocessor. For most ASR purposes "
        "AudioToMelSpectrogramPreprocessor does the same as the old "
        "AudioPreprocessing."
    )



# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Inspired by: https://github.com/r9y9/wavenet_vocoder
# Copyright (c) 2017: Ryuichi Yamamoto.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

import numpy as np
import torch
from torch.nn import functional as F


def dmld_loss(y_pred, y_true, num_classes):
    """Discretized mixture of logistic distributions loss

    https://github.com/r9y9/wavenet_vocoder/blob/master/wavenet_vocoder/mixture.py
    https://arxiv.org/pdf/1701.05517.pdf

    Args:
        y_pred (Tensor): Predicted output (B x T x C)
        y_true (Tensor): Target (B x T).
        num_classes (int): Number of classes

    Returns
        Tensor: loss

    """

    def log_sum_exp(x):
        """ numerically stable log_sum_exp implementation that prevents overflow """
        axis = len(x.size()) - 1
        m, _ = torch.max(x, dim=axis)
        m2, _ = torch.max(x, dim=axis, keepdim=True)
        return m + torch.log(torch.sum(torch.exp(x - m2), dim=axis))

    z_shape = y_pred.size(-1)
    assert z_shape % 3 == 0
    nr_mix = z_shape // 3

    # unpack parameters. (B, T, num_mixtures) x 3
    logit_probs = y_pred[:, :, :nr_mix]
    means = y_pred[:, :, nr_mix : 2 * nr_mix]
    log_scales = torch.clamp(y_pred[:, :, 2 * nr_mix : 3 * nr_mix], min=-7.0)

    # B x T -> B x T x num_mixtures
    y_true = y_true.unsqueeze(-1).expand_as(means)

    centered_y = y_true - means
    inv_stdv = torch.exp(-log_scales)
    plus_in = inv_stdv * (centered_y + 1.0 / (num_classes - 1))
    cdf_plus = torch.sigmoid(plus_in)
    min_in = inv_stdv * (centered_y - 1.0 / (num_classes - 1))
    cdf_min = torch.sigmoid(min_in)

    # log probability for edge case of 0 (before scaling)
    # equivalent: torch.log(torch.sigmoid(plus_in))
    log_cdf_plus = plus_in - F.softplus(plus_in)

    # log probability for edge case of 255 (before scaling)
    # equivalent: (1 - torch.sigmoid(min_in)).log()
    log_one_minus_cdf_min = -F.softplus(min_in)

    # probability for all other cases
    cdf_delta = cdf_plus - cdf_min

    mid_in = inv_stdv * centered_y
    # log probability in the center of the bin, to be used in extreme cases
    # (not actually used in our code)
    log_pdf_mid = mid_in - log_scales - 2.0 * F.softplus(mid_in)

    inner_inner_cond = (cdf_delta > 1e-5).float()
    # noinspection PyTypeChecker
    inner_inner_out = inner_inner_cond * torch.log(torch.clamp(cdf_delta, min=1e-12)) + (1.0 - inner_inner_cond) * (
        log_pdf_mid - np.log((num_classes - 1) / 2)
    )
    inner_cond = (y_true > 0.999).float()
    inner_out = inner_cond * log_one_minus_cdf_min + (1.0 - inner_cond) * inner_inner_out
    cond = (y_true < -0.999).float()
    log_probs = cond * log_cdf_plus + (1.0 - cond) * inner_out

    log_probs = log_probs + F.log_softmax(logit_probs, -1)

    return -log_sum_exp(log_probs)


def dmld_sample(y):
    """Sample from discretized mixture of logistic distributions.

    Args:
        y (Tensor): B x T x C

    Returns:
        Tensor: sample in range of [-1.0, 1.0].

    """

    z_shape = y.size(-1)
    assert z_shape % 3 == 0
    nr_mix = z_shape // 3

    # B x T x C
    logit_probs = y[:, :, :nr_mix]

    # sample mixture indicator from softmax
    temp = logit_probs.data.new(logit_probs.size()).uniform_(1e-5, 1.0 - 1e-5)
    temp = logit_probs.data - torch.log(-torch.log(temp))
    _, argmax = temp.max(dim=-1)

    # (B, T) -> (B, T, nr_mix)
    one_hot = torch.zeros(argmax.size() + (nr_mix,), dtype=torch.float, device=argmax.device)
    one_hot.scatter_(len(argmax.size()), argmax.unsqueeze(-1), 1.0)

    # select logistic parameters
    means = torch.sum(y[:, :, nr_mix : 2 * nr_mix] * one_hot, dim=-1)
    log_scales = torch.sum(y[:, :, 2 * nr_mix : 3 * nr_mix] * one_hot, dim=-1)
    log_scales = torch.clamp(log_scales, min=-7.0)
    # sample from logistic & clip to interval
    # we don't actually round to the nearest 8bit value when sampling
    u = means.data.new(means.size()).uniform_(1e-5, 1.0 - 1e-5)
    x = means + torch.exp(log_scales) * (torch.log(u) - torch.log(1.0 - u))

    x = torch.clamp(torch.clamp(x, min=-1.0), max=1.0)

    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os
import shutil
import tarfile
import unittest
from unittest import TestCase

import pytest
from ruamel.yaml import YAML

import nemo
import nemo.collections.asr as nemo_asr
from nemo.collections.asr.parts import AudioDataset, WaveformFeaturizer, collections, parsers
from nemo.core import DeviceType
from nemo.utils import logging

freq = 16000


@pytest.mark.usefixtures("neural_factory")
class TestUnitASRPytorch(TestCase):
    labels = [
        " ",
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
        "'",
    ]
    manifest_filepath = os.path.abspath(os.path.join(os.path.dirname(__file__), "../data/asr/an4_train.json"))
    featurizer_config = {
        'window': 'hann',
        'dither': 1e-05,
        'normalize': 'per_feature',
        'frame_splicing': 1,
        'int_values': False,
        'window_stride': 0.01,
        'sample_rate': freq,
        'features': 64,
        'n_fft': 512,
        'window_size': 0.02,
    }
    yaml = YAML(typ="safe")

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        data_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "../data/"))
        logging.info("Looking up for test ASR data")
        if not os.path.exists(os.path.join(data_folder, "asr")):
            logging.info("Extracting ASR data to: {0}".format(os.path.join(data_folder, "asr")))
            tar = tarfile.open(os.path.join(data_folder, "asr.tar.gz"), "r:gz")
            tar.extractall(path=data_folder)
            tar.close()
        else:
            logging.info("ASR data found in: {0}".format(os.path.join(data_folder, "asr")))

    @pytest.mark.unit
    def test_transcript_normalizers(self):
        # Create test json
        test_strings = [
            "TEST CAPITALIZATION",
            '!\\"#$%&\'()*+,-./:;<=>?@[\\\\]^_`{|}~',
            "3+3=10",
            "3 + 3 = 10",
            "why     is \\t whitepsace\\tsuch a problem   why indeed",
            "\\\"Can you handle quotes?,\\\" says the boy",
            "I Jump!!!!With joy?Now.",
            "Maybe I want to learn periods.",
            "$10 10.90 1-800-000-0000",
            "18000000000 one thousand 2020",
            "1 10 100 1000 10000 100000 1000000",
            "Î  ĻƠvɆȩȅĘ ÀÁÃ Ą ÇĊňńŤŧș",
            "‘’“”❛❜❝❞「 」 〈 〉 《 》 【 】 〔 〕 ⦗ ⦘ 😙  👀 🔨",
            "It only costs $1 000 000! Cheap right?",
            "2500, 3000 are separate but 200, 125 is not",
            "1",
            "1 2",
            "1 2 3",
            "10:00pm is 10:00 pm is 22:00 but not 10: 00 pm",
            "10:00 10:01pm 10:10am 10:90pm",
            "Mr. Expand me!",
            "Mr Don't Expand me!",
        ]
        normalized_strings = [
            "test capitalization",
            'percent and \' plus',
            "three plus three ten",
            "three plus three ten",
            "why is whitepsace such a problem why indeed",
            "can you handle quotes says the boy",
            "i jump with joy now",
            "maybe i want to learn periods",
            "ten dollars ten point nine zero one eight hundred zero zero",
            "eighteen billion one thousand two thousand and twenty",
            # Two line string below
            "one ten thousand one hundred one thousand ten thousand one hundred thousand one million",
            "i loveeee aaa a ccnntts",
            "''",
            "it only costs one million dollars cheap right",
            # Two line string below
            "two thousand five hundred three thousand are separate but two "
            "hundred thousand one hundred and twenty five is not",
            "one",
            "one two",
            "one two three",
            "ten pm is ten pm is twenty two but not ten zero pm",
            "ten ten one pm ten ten am ten ninety pm",
            "mister expand me",
            "mr don't expand me",
        ]
        manifest_paths = os.path.abspath(os.path.join(os.path.dirname(__file__), "../data/asr/manifest_test.json"))

        def remove_test_json():
            os.remove(manifest_paths)

        self.addCleanup(remove_test_json)

        with open(manifest_paths, "w") as f:
            for s in test_strings:
                f.write('{"audio_filepath": "", "duration": 1.0, "text": ' f'"{s}"}}\n')
        parser = parsers.make_parser(self.labels, 'en')
        manifest = collections.ASRAudioText(manifests_files=[manifest_paths], parser=parser,)

        for i, s in enumerate(normalized_strings):
            self.assertTrue(manifest[i].text_tokens == parser(s))

    @pytest.mark.unit
    def test_pytorch_audio_dataset(self):
        featurizer = WaveformFeaturizer.from_config(self.featurizer_config)
        ds = AudioDataset(manifest_filepath=self.manifest_filepath, labels=self.labels, featurizer=featurizer,)

        for i in range(len(ds)):
            if i == 5:
                logging.info(ds[i])
            # logging.info(ds[i][0].shape)
            # self.assertEqual(freq, ds[i][0].shape[0])

    @pytest.mark.unit
    def test_dataloader(self):
        batch_size = 4
        dl = nemo_asr.AudioToTextDataLayer(
            # featurizer_config=self.featurizer_config,
            manifest_filepath=self.manifest_filepath,
            labels=self.labels,
            batch_size=batch_size,
            # placement=DeviceType.GPU,
            drop_last=True,
        )
        for ind, data in enumerate(dl.data_iterator):
            # With num_workers update, this is no longer true
            # Moving to GPU is handled by AudioPreprocessor
            # data is on GPU
            # self.assertTrue(data[0].is_cuda)
            # self.assertTrue(data[1].is_cuda)
            # self.assertTrue(data[2].is_cuda)
            # self.assertTrue(data[3].is_cuda)
            # first dimension is batch
            self.assertTrue(data[0].size(0) == batch_size)
            self.assertTrue(data[1].size(0) == batch_size)
            self.assertTrue(data[2].size(0) == batch_size)
            self.assertTrue(data[3].size(0) == batch_size)

    @pytest.mark.unit
    def test_preprocessor_errors(self):
        def create_broken_preprocessor_1():
            nemo_asr.AudioToMelSpectrogramPreprocessor(window_size=2, n_window_size=2)

        def create_broken_preprocessor_2():
            nemo_asr.AudioToMelSpectrogramPreprocessor(window_stride=2, n_window_stride=2)

        def create_broken_preprocessor_3():
            nemo_asr.AudioToMelSpectrogramPreprocessor(n_window_stride=2)

        def create_good_preprocessor_1():
            nemo_asr.AudioToMelSpectrogramPreprocessor(window_size=0.02, window_stride=0.01)

        def create_good_preprocessor_2():
            nemo_asr.AudioToMelSpectrogramPreprocessor(
                window_size=None, window_stride=None, n_window_size=256, n_window_stride=32,
            )

        self.assertRaises(ValueError, create_broken_preprocessor_1)
        self.assertRaises(ValueError, create_broken_preprocessor_2)
        self.assertRaises(ValueError, create_broken_preprocessor_3)
        create_good_preprocessor_1()
        create_good_preprocessor_2()

    @pytest.mark.unit
    def test_kaldi_dataloader(self):
        batch_size = 4
        dl = nemo_asr.KaldiFeatureDataLayer(
            kaldi_dir=os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/asr/kaldi_an4/')),
            labels=self.labels,
            batch_size=batch_size,
        )
        for data in dl.data_iterator:
            self.assertTrue(data[0].size(0) == batch_size)

        dl_test_min = nemo_asr.KaldiFeatureDataLayer(
            kaldi_dir=os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/asr/kaldi_an4/')),
            labels=self.labels,
            batch_size=batch_size,
            min_duration=1.0,
        )
        self.assertTrue(len(dl_test_min) == 18)

        dl_test_max = nemo_asr.KaldiFeatureDataLayer(
            kaldi_dir=os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/asr/kaldi_an4/')),
            labels=self.labels,
            batch_size=batch_size,
            max_duration=5.0,
        )
        self.assertTrue(len(dl_test_max) == 19)

    @pytest.mark.unit
    def test_tarred_dataloader(self):
        batch_size = 4
        manifest_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '../data/asr/tarred_an4/tarred_audio_manifest.json')
        )

        # Test loading a single tarball
        tarpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/asr/tarred_an4/audio_0.tar'))
        dl_single_tar = nemo_asr.TarredAudioToTextDataLayer(
            audio_tar_filepaths=tarpath, manifest_filepath=manifest_path, labels=self.labels, batch_size=batch_size
        )
        count = 0
        for _ in dl_single_tar.dataset:
            count += 1
        self.assertTrue(count == 16)

        # Test braceexpand loading
        tarpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '../data/asr/tarred_an4/audio_{0..3}.tar'))
        dl_braceexpand = nemo_asr.TarredAudioToTextDataLayer(
            audio_tar_filepaths=tarpath, manifest_filepath=manifest_path, labels=self.labels, batch_size=batch_size
        )
        self.assertTrue(len(dl_braceexpand) == 65)
        count = 0
        for _ in dl_braceexpand.dataset:
            count += 1
        self.assertTrue(count == 65)

        # Test loading via list
        tarpath = [
            os.path.abspath(os.path.join(os.path.dirname(__file__), f'../data/asr/tarred_an4/audio_{i}.tar'))
            for i in range(4)
        ]
        dl_list_load = nemo_asr.TarredAudioToTextDataLayer(
            audio_tar_filepaths=tarpath, manifest_filepath=manifest_path, labels=self.labels, batch_size=batch_size
        )
        count = 0
        for _ in dl_braceexpand.dataset:
            count += 1
        self.assertTrue(count == 65)

    @pytest.mark.unit
    def test_trim_silence(self):
        batch_size = 4
        normal_dl = nemo_asr.AudioToTextDataLayer(
            # featurizer_config=self.featurizer_config,
            manifest_filepath=self.manifest_filepath,
            labels=self.labels,
            batch_size=batch_size,
            # placement=DeviceType.GPU,
            drop_last=True,
            shuffle=False,
        )
        trimmed_dl = nemo_asr.AudioToTextDataLayer(
            # featurizer_config=self.featurizer_config,
            manifest_filepath=self.manifest_filepath,
            trim_silence=True,
            labels=self.labels,
            batch_size=batch_size,
            # placement=DeviceType.GPU,
            drop_last=True,
            shuffle=False,
        )
        for norm, trim in zip(normal_dl.data_iterator, trimmed_dl.data_iterator):
            for point in range(batch_size):
                self.assertTrue(norm[1][point].data >= trim[1][point].data)

    @pytest.mark.unit
    def test_audio_preprocessors(self):
        batch_size = 5
        dl = nemo_asr.AudioToTextDataLayer(
            # featurizer_config=self.featurizer_config,
            manifest_filepath=self.manifest_filepath,
            labels=self.labels,
            batch_size=batch_size,
            # placement=DeviceType.GPU,
            drop_last=True,
            shuffle=False,
        )

        installed_torchaudio = True
        try:
            import torchaudio
        except ModuleNotFoundError:
            installed_torchaudio = False
            with self.assertRaises(ModuleNotFoundError):
                to_spectrogram = nemo_asr.AudioToSpectrogramPreprocessor(n_fft=400, window=None)
            with self.assertRaises(ModuleNotFoundError):
                to_mfcc = nemo_asr.AudioToMFCCPreprocessor(n_mfcc=15)

        if installed_torchaudio:
            to_spectrogram = nemo_asr.AudioToSpectrogramPreprocessor(n_fft=400, window=None)
            to_mfcc = nemo_asr.AudioToMFCCPreprocessor(n_mfcc=15)

        to_melspec = nemo_asr.AudioToMelSpectrogramPreprocessor(features=50)

        for batch in dl.data_iterator:
            input_signals, seq_lengths, _, _ = batch
            input_signals = input_signals.to(to_melspec._device)
            seq_lengths = seq_lengths.to(to_melspec._device)

            melspec = to_melspec.forward(input_signals, seq_lengths)

            if installed_torchaudio:
                spec = to_spectrogram.forward(input_signals, seq_lengths)
                mfcc = to_mfcc.forward(input_signals, seq_lengths)

            # Check that number of features is what we expect
            self.assertTrue(melspec[0].shape[1] == 50)

            if installed_torchaudio:
                self.assertTrue(spec[0].shape[1] == 201)  # n_fft // 2 + 1 bins
                self.assertTrue(mfcc[0].shape[1] == 15)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/blob/master/convlab2/util/multiwoz/dbquery.py
"""
import json
import os
import random

__all__ = ['Database']


class Database(object):
    def __init__(self, data_dir):
        super(Database, self).__init__()
        # loading databases
        domains = ['restaurant', 'hotel', 'attraction', 'train', 'hospital', 'taxi', 'police']
        self.dbs = {}
        for domain in domains:
            with open(os.path.join(data_dir, 'db/{}_db.json'.format(domain))) as f:
                self.dbs[domain] = json.load(f)

    def query(self, domain, constraints, ignore_open=True):
        """Returns the list of entities for a given domain
        based on the annotation of the belief state"""
        # query the db
        if domain == 'taxi':
            return [
                {
                    'taxi_colors': random.choice(self.dbs[domain]['taxi_colors']),
                    'taxi_types': random.choice(self.dbs[domain]['taxi_types']),
                    'taxi_phone': ''.join([str(random.randint(1, 9)) for _ in range(11)]),
                }
            ]
        if domain == 'police':
            return self.dbs['police']
        if domain == 'hospital':
            return self.dbs['hospital']

        found = []
        for i, record in enumerate(self.dbs[domain]):
            for key, val in constraints:
                if (
                    val == ""
                    or val == "dont care"
                    or val == 'not mentioned'
                    or val == "don't care"
                    or val == "dontcare"
                    or val == "do n't care"
                ):
                    pass
                else:
                    try:
                        record_keys = [k.lower() for k in record]
                        if key.lower() not in record_keys:
                            continue
                        if key == 'leaveAt':
                            val1 = int(val.split(':')[0]) * 100 + int(val.split(':')[1])
                            val2 = int(record['leaveAt'].split(':')[0]) * 100 + int(record['leaveAt'].split(':')[1])
                            if val1 > val2:
                                break
                        elif key == 'arriveBy':
                            val1 = int(val.split(':')[0]) * 100 + int(val.split(':')[1])
                            val2 = int(record['arriveBy'].split(':')[0]) * 100 + int(record['arriveBy'].split(':')[1])
                            if val1 < val2:
                                break
                        # elif ignore_open and key in ['destination', 'departure', 'name']:
                        elif ignore_open and key in ['destination', 'departure']:
                            continue
                        else:
                            if val.strip() != record[key].strip():
                                break
                    except Exception:
                        continue
            else:
                record['Ref'] = '{0:08d}'.format(i)
                found.append(record)

        return found
